from packyak.spec import ResourceType


class Resource:
    resource_type: ResourceType

    def __init__(self, resource_type: ResourceType, resource_id: str):
        self.resource_type = resource_type
        self.resource_id = resource_id
        self.bucket_id = self.resource_id
        resources.append(self)


resources: list[Resource] = []



from __future__ import annotations

from typing import Literal, Optional
from pydantic import BaseModel

type BucketSubscriptionScope = Literal["create"] | Literal["update"] | Literal["delete"]

type ResourceType = Literal["bucket"] | Literal["queue"] | Literal["function"]


class BindingSpec(BaseModel):
    resource_type: ResourceType
    resource_id: str
    scopes: list[str]
    props: dict[str, str] | None


class BucketBindingSpec(BindingSpec):
    selector: Optional[str]


class FunctionSpec(BaseModel):
    function_id: str
    file_name: str
    bindings: list[BindingSpec]


class BucketSubscriptionSpec(BaseModel):
    scope: BucketSubscriptionScope
    function_id: str


class BucketSpec(BaseModel):
    bucket_id: str
    subscriptions: list[BucketSubscriptionSpec]


class QueueSubscriptionSpec(BaseModel):
    function_id: str


class QueueSpec(BaseModel):
    queue_id: str
    fifo: bool
    subscriptions: list[QueueSubscriptionSpec]


class PackyakSpec(BaseModel):
    buckets: list[BucketSpec]
    queues: list[QueueSpec]
    functions: list[FunctionSpec]



from __future__ import annotations

import json
import os
from typing import Any, Callable, Type, cast

import aioboto3
import boto3
from types_aiobotocore_sqs.type_defs import ReceiveMessageResultTypeDef

from pydantic import BaseModel
from types_aiobotocore_sqs import SQSClient

from .function import LambdaFunction, function
from .globals import QUEUES
from .integration import integration
from .resource import Resource
from .typed_resource import TypedResource

sqs = boto3.client("sqs")
session: aioboto3.Session = aioboto3.Session()

aio_sqs = TypedResource[SQSClient](session.client("sqs"))  # type: ignore

type Body = str | BaseModel


class ReceivedMessagesEvent[B: Body]:
    def __init__(self, messages: list[Message[B]]):
        self.messages = messages


class Message[B: Body]:
    def __init__(self, queue: "Queue[B]", message_id: str, body: B):
        self.queue = queue
        self.message_id = message_id
        self.body = body

    async def delete(self) -> None:
        await self.queue.delete(receipt_handle=self.message_id)


class QueueSubscription:
    def __init__(
        self,
        queue: "Queue[Any]",
        function: LambdaFunction[Any, Any],
    ):
        self.queue = queue
        self.function = function


class Queue[B: Body](Resource):
    subscriptions: list[QueueSubscription]

    def __init__(self, resource_id: str, model: Type[B] = str, fifo: bool = False):
        super().__init__("queue", resource_id)
        self.model = model
        self.fifo = fifo
        self.subscriptions = []
        if resource_id in QUEUES:
            raise Exception(f"Queue {resource_id} already exists")
        QUEUES[resource_id] = self

    @property
    def queue_url(self):
        queue_url_env = f"{self.resource_id}_queue_url"
        queue_url = os.getenv(queue_url_env)
        if queue_url is None:
            raise ValueError(f"{queue_url_env} is not defined")
        return queue_url

    def send_sync(self, body: Body) -> None:
        sqs.send_message(QueueUrl=self.queue_url, MessageBody=to_json(body))

    @integration("write")
    async def send(self, body: Body) -> None:
        async with aio_sqs as sqs:
            await sqs.send_message(QueueUrl=self.queue_url, MessageBody=to_json(body))

    def receive_sync(self) -> list[Message[B]]:
        return self._parse_receive(sqs.receive_message(QueueUrl=self.queue_url))

    @integration("read")
    async def receive(self) -> list[Message[B]]:
        async with aio_sqs as sqs:
            return self._parse_receive(
                await sqs.receive_message(QueueUrl=self.queue_url)
            )

    def _parse_receive(self, response: ReceiveMessageResultTypeDef) -> list[Message[B]]:
        return [
            Message(self, msg["MessageId"], self.from_json(msg["Body"]))  # type: ignore
            for msg in response.get("Messages", [])
        ]

    @integration("delete")
    async def delete(self, receipt_handle: str) -> None:
        async with aio_sqs as sqs:
            await sqs.delete_message(
                QueueUrl=self.queue_url, ReceiptHandle=receipt_handle
            )

    @integration("list")
    async def list(self):
        async with aio_sqs as sqs:
            response = await sqs.list_queues()
            return response.get("QueueUrls", [])

    @integration("change_visibility")
    async def change_visibility(self, receipt_handle: str, visibility_timeout: int):
        async with aio_sqs as sqs:
            await sqs.change_message_visibility(
                QueueUrl=self.queue_url,
                ReceiptHandle=receipt_handle,
                VisibilityTimeout=visibility_timeout,
            )

    def consumer(self, function_id: str | None = None):
        def decorate(handler: Callable[[ReceivedMessagesEvent[B]], Any]):
            from aws_lambda_typing.events.sqs import SQSEvent

            # see https://kevinhakanson.com/2022-04-10-python-typings-for-aws-lambda-function-events/
            @function(function_id=function_id or handler.__name__)
            async def lambda_func(event: SQSEvent, context: Any):
                result = await handler(
                    ReceivedMessagesEvent(
                        messages=[
                            Message(
                                queue=self,
                                message_id=message["messageId"],
                                body=self.from_json(message["body"]),
                            )
                            for message in event["Records"]
                        ]
                    )
                )
                # TODO: adapt response
                return result

            self.subscriptions.append(QueueSubscription(self, lambda_func))

        return decorate

    def from_json(self, body: str) -> B:
        if issubclass(self.model, BaseModel):
            return cast(B, self.model.model_validate_json(body))  # type: ignore
        else:
            return json.loads(body)


def to_json(body: Body):
    if isinstance(body, BaseModel):
        return body.model_dump_json()
    else:
        return str(body)



from typing import (
    Callable,
    Generic,
    Optional,
    TypeVar,
    Protocol,
    ParamSpec,
    runtime_checkable,
    cast,
)

from packyak.globals import FUNCTIONS

Params = ParamSpec("Params")
Return = TypeVar("Return", covariant=True)


@runtime_checkable
class LambdaFunction(Protocol, Generic[Params, Return]):
    file_name: str
    function_id: str

    def __call__(self, *args: Params.args, **kwargs: Params.kwargs) -> Return:
        ...


def function(*, function_id: Optional[str] = None):
    def decorator(func: Callable[Params, Return]) -> LambdaFunction[Params, Return]:
        # @wraps(func)
        def wrapper(*args: Params.args, **kwargs: Params.kwargs) -> Return:
            return func(*args, **kwargs)

        func_id = function_id or func.__name__
        setattr(wrapper, "file_name", func.__code__.co_filename)
        setattr(wrapper, "function_id", func_id)

        if func_id in FUNCTIONS:
            raise Exception(f"Lambda Function {func_id} already exists")

        lambda_func = cast(LambdaFunction[Params, Return], wrapper)
        FUNCTIONS[func_id] = lambda_func
        return lambda_func

    return decorator



import ast
from functools import partial
import inspect
from typing import Any, Callable, Sequence, cast

from packyak.globals import BUCKETS, FUNCTIONS, QUEUES

from packyak.resource import Resource

from .binding import Binding
from .function import LambdaFunction
from .integration import Integration


def find_all_resources() -> Sequence[Resource]:
    return list(BUCKETS.values()) + list(QUEUES.values())


def find_all_functions() -> list[LambdaFunction[Any, Any]]:
    return list(FUNCTIONS.values())


type Func = Callable[..., Any]


# walks a function to detect its dependencies
def find_bindings(func: LambdaFunction[Any, Any]) -> list[Binding]:
    print("analyze", func.function_id)
    bindings: list[Binding] = []

    for call in find_calls(func):
        if call.obj is None:
            raise Exception("Integration must be called on an object")
        bindings.append(Binding(func, call.obj, call.func.scopes, call.metadata))

    return bindings


class Call:
    def __init__(
        self,
        func: Integration[Any, Any],
        *,
        obj: Any = None,
        metadata: dict[str, property] | None = None,
    ):
        self.func = func
        self.obj = obj
        self.metadata = metadata


def find_calls(
    func: Func, seen: set[Func] = set(), *, obj: object | None = None
) -> list[Call]:
    if func in seen:
        return []

    seen.add(func)

    if hasattr(func, "scopes") and obj is not None:
        # this is an Integration
        integration = cast(Integration[Any, Any], func)
        metadata = {}
        if integration.metadata is not None:
            for k, v in integration.metadata.items():
                if isinstance(v, property):
                    metadata[k] = getattr(obj, v.fget.__name__)
                else:
                    metadata[k] = getattr(obj, v)

        return [
            Call(
                func=func,  # type: ignore
                obj=obj,
                metadata=metadata,
            )
        ]
    elif isinstance(func, partial):
        print("partial")
        return find_calls(func.func, seen, obj=obj)

    lexical_scope = get_lexical_scope(func)

    def lookup(expr: str):
        return eval(expr, lexical_scope)

    # Get the source of the function
    try:
        src = dedent(inspect.getsource(func))
    except TypeError:
        # TODO: Handle builtins
        return []

    # Parse the source into an AST
    module = ast.parse(src)

    # Custom visitor class to find function calls
    class CallVisitor(ast.NodeVisitor):
        calls: list[Call]

        def __init__(self):
            self.calls = []

        def visit_Call(self, node: ast.Call):
            try:
                if isinstance(node.func, ast.Name):
                    self.calls.extend(find_calls(lookup(node.func.id), seen))
                elif isinstance(node.func, ast.Attribute) and isinstance(
                    node.func.value, ast.Name
                ):
                    calls = find_calls(
                        lookup(f"{node.func.value.id}.{node.func.attr}"),
                        seen,
                        obj=lookup(node.func.value.id),
                    )
                    self.calls.extend(calls)

                # elif isinstance(node.func, ast.Subscript):
                #     call = Call(
                #         lookup(f"{node.func.value.id}[{node.func.slice.value.s!r}]"),
                #         lookup(f"{node.func.value.id}"),
                #     )
                #     self.calls.append(call)
            except NameError:
                pass

    visitor = CallVisitor()
    visitor.visit(module)
    return visitor.calls


# Get the lexical scope of a function
def get_lexical_scope(func: Func) -> dict[str, Any]:
    if isinstance(func, partial):
        return get_lexical_scope(func.func)

    globals = func.__globals__ if hasattr(func, "__globals__") else {}
    closure_vars = {}
    if hasattr(func, "__closure__") and func.__closure__ is not None:
        closure_var_names = func.__code__.co_freevars
        closure_vars = {
            name: cell.cell_contents
            for name, cell in zip(closure_var_names, func.__closure__)
        }

    lexical_scope = {**globals, **closure_vars}

    return lexical_scope


def dedent(src: str) -> str:
    lines = src.split("\n")
    # Filter out blank lines and find the minimum indentation
    min_indent = min(len(line) - len(line.lstrip()) for line in lines if line.strip())
    # Remove the minimum indentation from all lines
    dedented_lines = [line[min_indent:] if line.strip() else line for line in lines]
    return "\n".join(dedented_lines)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .bucket import Bucket
    from .function import LambdaFunction
    from .queue import Queue

BUCKETS: dict[str, Bucket] = {}
FUNCTIONS: dict[str, LambdaFunction[Any, Any]] = {}
QUEUES: dict[str, Queue[Any]] = {}



import os
from typing import Any
from .bucket import Bucket
from .function import LambdaFunction
from .queue import Queue
from .reflect import find_all_functions, find_all_resources, find_bindings
from .resource import Resource
from .spec import (
    BindingSpec,
    BucketSpec,
    BucketSubscriptionSpec,
    FunctionSpec,
    QueueSpec,
    QueueSubscriptionSpec,
    PackyakSpec,
)


def is_synth():
    packyak_synth = os.environ.get("PACKYAK_SYNTH")
    return packyak_synth is not None and (
        packyak_synth == "1" or packyak_synth.lower() == "true"
    )


def synth() -> PackyakSpec:
    functions: list[FunctionSpec] = []
    buckets: list[BucketSpec] = []
    queues: list[QueueSpec] = []

    seen = set[Any]()

    def visit(resource: Resource | LambdaFunction[Any, Any]):
        if resource in seen:
            return
        seen.add(resource)

        if isinstance(resource, Bucket):
            buckets.append(
                BucketSpec(
                    bucket_id=resource.resource_id,
                    subscriptions=[
                        BucketSubscriptionSpec(
                            scope=sub.scope,  # type: ignore - sub.scope is typed properly
                            function_id=sub.function.function_id,
                        )
                        for sub in resource.subscriptions
                    ],
                )
            )
        elif isinstance(resource, Queue):
            queues.append(
                QueueSpec(
                    queue_id=resource.resource_id,
                    fifo=resource.fifo,
                    subscriptions=[
                        QueueSubscriptionSpec(function_id=sub.function.function_id)
                        for sub in resource.subscriptions
                    ],
                )
            )
        elif isinstance(resource, LambdaFunction):
            functions.append(
                FunctionSpec(
                    function_id=resource.function_id,
                    file_name=resource.file_name,
                    bindings=[
                        BindingSpec(
                            resource_type=binding.resource.resource_type,
                            resource_id=binding.resource.resource_id,
                            scopes=binding.scopes,
                            props=binding.metadata,
                        )
                        for binding in find_bindings(resource)
                    ],
                )
            )

    for resource in find_all_resources():
        visit(resource)

    for function in find_all_functions():
        visit(function)

    packyak_spec = PackyakSpec(
        buckets=buckets,
        queues=queues,
        functions=functions,
    )
    return packyak_spec



from asyncio import AbstractEventLoop, get_event_loop, new_event_loop
from typing import Any, Coroutine, TypeVar

T = TypeVar("T")


def sync(coroutine: Coroutine[Any, Any, T]) -> T:
    loop: AbstractEventLoop = get_event_loop()
    if loop.is_closed():
        loop = new_event_loop()
    return loop.run_until_complete(coroutine)



from datetime import datetime
import os
from typing import Any, Callable, Optional, cast

import aioboto3
import boto3
from botocore.response import StreamingBody
from pydantic import BaseModel
from types_aiobotocore_s3.client import S3Client
from types_aiobotocore_s3.type_defs import ListObjectsV2OutputTypeDef

from packyak.typed_resource import TypedResource

from .folder import Folder
from .function import LambdaFunction, function
from .globals import BUCKETS
from .integration import integration
from .resource import Resource
from .spec import BucketSubscriptionScope
from .sync import sync

s3 = boto3.client("s3")
session: aioboto3.Session = aioboto3.Session()

aio_s3 = TypedResource[S3Client](session.client("s3"))  # type: ignore


class ObjectRef:
    def __init__(self, bucket: "Bucket", key: str, etag: str):
        self.bucket = bucket
        self.key = key
        self.etag = etag
        self.created_at = datetime.now()

    def __hash__(self):
        return hash((self.bucket.bucket_id, self.key, self.etag, self.created_at))


class Object[Body](ObjectRef):
    def __init__(self, bucket: "Bucket", key: str, body: Body, etag: str):
        super().__init__(bucket, key, etag)
        self.body = body


class ListObjectsResponse:
    def __init__(
        self, bucket: "Bucket", objects: list[ObjectRef], next_token: str | None
    ):
        self.bucket = bucket
        self.objects = objects
        self.next_token = next_token

    def __aiter__(self):
        return self

    async def __anext__(self):
        pass


class BucketSubscription:
    def __init__(
        self,
        bucket: "Bucket",
        scope: BucketSubscriptionScope,
        function: LambdaFunction[Any, Any],
        prefix: str | None = None,
    ):
        self.bucket = bucket
        self.scope = scope
        self.function = function
        self.prefix = prefix


class Bucket(Resource):
    subscriptions: list[BucketSubscription]

    def __init__(self, bucket_id: str):
        super().__init__("bucket", resource_id=bucket_id)
        self.subscriptions = []
        if self.resource_id in BUCKETS:
            raise Exception(f"Bucket {self.resource_id} already exists")
        BUCKETS[self.resource_id] = self

    class Event(BaseModel):
        key: str
        etag: str

    class ObjectCreatedEvent(Event):
        size: int | None

    class ObjectDeletedEvent(Event):
        pass

    # Overload the '/' operator
    def __truediv__(self, name: str) -> Folder:
        return self.folder(name)

    def folder(self, name: str) -> Folder:
        _folder: Optional[Folder] = None
        for path_component in name.split("/"):
            if _folder is None:
                _folder = Folder(self, path_component)
            else:
                _folder = _folder / path_component
        return cast(Folder, _folder)

    @property
    def bucket_name(self):
        bucket_name_env = f"{self.resource_id}_bucket_name"
        bucket_name = os.getenv(bucket_name_env)
        if bucket_name is None:
            raise ValueError(f"{bucket_name_env} is not defined")
        return bucket_name

    def put_sync(self, key: str, body: str):
        return sync(self.put(key, body))

    @integration("write")
    async def put(self, key: str, body: str):
        async with aio_s3 as s3:
            await s3.put_object(Bucket=self.bucket_name, Key=key, Body=body)

    @integration("read")
    def get_sync(self, key: str) -> Object[StreamingBody]:
        response = s3.get_object(Bucket=self.bucket_name, Key=key)
        return Object(
            bucket=self,
            key=key,
            body=response["Body"],
            etag=response["ETag"],
        )

    @integration("read")
    async def get(self, key: str) -> Object[StreamingBody]:
        async with aio_s3 as s3:
            response = await s3.get_object(Bucket=self.bucket_name, Key=key)
            return Object(
                bucket=self,
                key=key,
                body=response["Body"],
                etag=response["ETag"],
            )

    @integration("delete")
    def delete_sync(self, key: str):
        s3.delete_object(Bucket=self.bucket_name, Key=key)

    @integration("delete")
    async def delete(self, key: str):
        async with aio_s3 as s3:
            await s3.delete_object(Bucket=self.bucket_name, Key=key)

    @integration("list")
    def list_sync(
        self, prefix: str, *, limit: int | None, next_token: Optional[str] = None
    ):
        response = s3.list_objects_v2(
            Bucket=self.bucket_name,
            MaxKeys=limit or 100,
            ContinuationToken=cast(str, next_token or None),
            Prefix=prefix,
        )
        return self.parse_list_objects_v2_response(response)

    @integration("list")
    async def list(
        self, prefix: str, *, limit: int | None, next_token: Optional[str] = None
    ):
        async with aio_s3 as client:
            response = await client.list_objects_v2(
                Bucket=self.bucket_name,
                MaxKeys=limit or 100,
                ContinuationToken=cast(str, next_token or None),
                Prefix=prefix,
            )
            return self.parse_list_objects_v2_response(response)

    def parse_list_objects_v2_response(
        self, response: ListObjectsV2OutputTypeDef
    ) -> ListObjectsResponse:
        return ListObjectsResponse(
            bucket=self,
            objects=[
                ObjectRef(self, obj_ref["Key"], obj_ref["ETag"])  # type: ignore
                for obj_ref in response.get("Contents", [])
            ],
            next_token=response.get("NextContinuationToken"),
        )

    def on(
        self,
        scope: BucketSubscriptionScope,
        prefix: Optional[str] = None,
        function_id: str | None = None,
    ):
        def decorate(handler: Callable[[Bucket.ObjectCreatedEvent], Any]):
            from aws_lambda_typing.events.event_bridge import EventBridgeEvent
            from aws_lambda_typing.events.s3 import S3

            # see https://kevinhakanson.com/2022-04-10-python-typings-for-aws-lambda-function-events/
            @function(function_id=function_id or handler.__name__)
            async def lambda_func(event: EventBridgeEvent, context: Any):
                event_detail = S3(event["detail"])  # type: ignore
                result = await handler(
                    Bucket.ObjectCreatedEvent(
                        key=event_detail["object"]["key"],  # type: ignore
                        etag=event_detail["object"]["etag"],  # type: ignore
                        size=event_detail["object"]["size"],  # type: ignore
                    )
                )
                return result

            self.subscriptions.append(
                BucketSubscription(self, scope, lambda_func, prefix)
            )

            return lambda_func

        return decorate



import os
from typing import Never
from .function import function
from .bucket import Bucket
from .queue import Queue, Message, ReceivedMessagesEvent
from .synth import synth, is_synth


def init() -> None | Never:
    """
    Synthesizes the packyak spec to disk if the PACKYAK_SYNTH environment variable is set.

    If PACKYAK_SYNTH is falsy, then this function does nothing.

    Otherwise, it analyzes the application, writes the spec to .packyak/spec.json, and exits.

    Why it exits:
    This function is designed to be called at the end of a program or before a script.
    └→ e.g. for streamlit scripts, we don't want to run the streamlit application during build, only during runtime.
    """
    if is_synth():
        packyak_spec = synth()
        if not os.path.exists(".packyak"):
            os.makedirs(".packyak")
        with open(".packyak/spec.json", "w") as f:
            f.write(
                packyak_spec.model_dump_json(
                    indent=2, exclude_unset=True, exclude_none=True
                )
            )
        exit(0)
    else:
        # TODO: initialize runtime environment variables
        pass


__all__ = ["function", "Bucket", "Queue", "Message", "ReceivedMessagesEvent", "init"]



from typing import Any

from .function import LambdaFunction
from .resource import Resource


class Binding:
    def __init__(
        self,
        function: LambdaFunction[Any, Any],
        resource: Resource,
        scopes: list[str],
        metadata: dict[str, Any] | None = None,
    ) -> None:
        self.resource = resource
        self.scopes = scopes
        self.function = function
        self.metadata = metadata
